/*
 * SHMACGrid.cpp
 *
 *  Created on: 7 Mar 2013
 *      Author: Yaman Umuroglu
 */

#include "SHMACGrid.h"

// construct a NULL-initialised grid of given dimensions
SHMACGrid::SHMACGrid(uint8_t rowCount, uint8_t colCount)
{
	sc_assert(rowCount > 0);
	sc_assert(colCount > 0);

	m_rowCount = rowCount;
	m_colCount = colCount;

	// construct grid of given size with NULL elements
	std::vector<SHMACTile *> row;
	for (uint8_t i = 0; i < m_rowCount; i++)
	{
		row.clear();
		for (uint8_t j = 0; j < m_colCount; j++)
		{
			row.push_back((SHMACTile*) NULL);
		}
		m_elements.push_back(row);
	}
}

SHMACGrid::~SHMACGrid()
{
	// delete each element in grid
	for (uint8_t i = 0; i < m_rowCount; i++)
		for (uint8_t j = 0; j < m_colCount; j++)
			if (m_elements[i][j] != NULL)
				delete m_elements[i][j];

}

SHMACTile* SHMACGrid::get(uint8_t row, uint8_t col) const
{
	// assert validity of coordinates
	sc_assert(row >= 0);
	sc_assert(row < m_rowCount);
	sc_assert(col >= 0);
	sc_assert(col < m_colCount);

	return m_elements.at(row).at(col);
}

SHMACTile* SHMACGrid::get(shmacsim::tile_coord coord) const
{
	return this->get(coord.i, coord.j);
}

void SHMACGrid::set(uint8_t row, uint8_t col, SHMACTile* setTo)
{
	// assert validity of coordinates
	sc_assert(row >= 0);
	sc_assert(row < m_rowCount);
	sc_assert(col >= 0);
	sc_assert(col < m_colCount);

	m_elements[row][col] = setTo;
}

void SHMACGrid::set(shmacsim::tile_coord coord, SHMACTile* setTo)
{
	this->set(coord.i, coord.j, setTo);
}

// convenience function to establish all tile connections defined in this SHMACGrid
void SHMACGrid::establishConnections()
{
	connectGrid(this, m_rowCount, m_colCount);
}

ostream& operator<<(ostream& file, const SHMACGrid& dbv)
{
	file << "SHMACGrid output: " << std::endl;
	for (uint8_t i = 0; i < dbv.m_rowCount; i++)
	{
		for (uint8_t j = 0; j < dbv.m_colCount; j++)
		{
			file << std::hex << dbv.get(i, j) << std::dec << "\t";
		}
		file << std::endl;
	}
	return (file << std::endl);
}

// convenience function for connecting two neighbouring SHMACTile's
// should not be called more than once, as SystemC ports don't like being bound twice
void SHMACGrid::connectNeighbors(SHMACTile* tile1, SHMACTile* tile2)
{
	tile1->connectToTile(tile2);
	tile2->connectToTile(tile1);
	/*

	// make sure we are operating on valid tiles
	sc_assert(tile1);
	sc_assert(tile2);

	shmacsim::tile_coord c1 = tile1->getCoords();
	shmacsim::tile_coord c2 = tile2->getCoords();
	int di = c1.i - c2.i, dj = c1.j - c2.j;

	// tiles to be connected must be horizontal or vertical neighbours
	if (abs(di) + abs(dj) != 1)
	{
		std::cerr
				<< "SHMACTile::connectNeighbors used on non-neighbouring tiles!"
				<< std::endl;
		sc_assert(0);
	}

	SHMACRouter * router1 = tile1->getRouter();
	SHMACRouter * router2 = tile2->getRouter();

	// handle neighbour connections on a case-by-case basis
	// each case with tile positions is illustrated
	if (di == -1 && dj == 0)
	{
		// case:	xx xx xx
		//       	xx c1 xx
		//       	xx c2 xx
		//			xx xx xx
		router1->routerPorts[shmacsim::DIR_SOUTH].bind(*router2);
		router2->routerPorts[shmacsim::DIR_NORTH].bind(*router1);
	}
	else if (di == 1 && dj == 0)
	{
		// case:	xx xx xx
		//       	xx c2 xx
		//       	xx c1 xx
		//			xx xx xx
		router1->routerPorts[shmacsim::DIR_NORTH].bind(*router2);
		router2->routerPorts[shmacsim::DIR_SOUTH].bind(*router1);
	}
	else if (di == 0 && dj == -1)
	{
		// case:	xx xx xx
		//       	xx c1 c2
		//       	xx xx xx
		//			xx xx xx
		router1->routerPorts[shmacsim::DIR_EAST].bind(*router2);
		router2->routerPorts[shmacsim::DIR_WEST].bind(*router1);
	}
	else if (di == 0 && dj == 1)
	{
		// case:	xx xx xx
		//       	xx c2 c1
		//       	xx xx xx
		//			xx xx xx
		router1->routerPorts[shmacsim::DIR_WEST].bind(*router2);
		router2->routerPorts[shmacsim::DIR_EAST].bind(*router1);
	}
	else
	{
		std::cerr
				<< "SHMACTile::connectNeighbors: Unrecognised neighbour placement!"
				<< std::endl;
		sc_assert(0);
	}
	*/
}

// convenience function for making connections in a SHMACGrid of given dimensions
void SHMACGrid::connectGrid(SHMACGrid * grid, uint8_t rowCount, uint8_t colCount)
{
	sc_assert(rowCount);
	sc_assert(colCount);

	// iterate over grid and connect tiles to east and south
	for (uint8_t i = 0; i < rowCount; i++)
		for (uint8_t j = 0; j < colCount; j++)
		{
			if (i + 1 < rowCount)
				connectNeighbors(grid->get(i, j), grid->get(i + 1, j));
			if (j + 1 < colCount)
				connectNeighbors(grid->get(i, j), grid->get(i, j + 1));
		}
}

SHMACTile*& SHMACGrid::operator ()(const uint8_t row, const uint8_t col)
{
	return m_elements[row][col];
}

void SHMACGrid::callRuntimeInitializers()
{
	// iterate over grid and call the runtime initializer for each tile
	for (uint8_t i = 0; i < m_rowCount; i++)
		for (uint8_t j = 0; j < m_colCount; j++)
			m_elements[i][j]->executeRuntimeInitializer();
}

void SHMACGrid::setMetaDataString(std::string key, std::string value)
{
	// iterate over grid and call metadata setter for each tile
	for (uint8_t i = 0; i < m_rowCount; i++)
		for (uint8_t j = 0; j < m_colCount; j++)
			m_elements[i][j]->setMetaDataString(key, value);
}

void SHMACGrid::setMetaDataInt(std::string key, unsigned int value)
{
	// iterate over grid and call metadata setter for each tile
	for (uint8_t i = 0; i < m_rowCount; i++)
		for (uint8_t j = 0; j < m_colCount; j++)
			m_elements[i][j]->setMetaDataInt(key, value);
}
