/*
 * SHMACLibrary
 *
 * Acts as a "factory" for SHMAC tiles (tile types must be registered first)
 * It essentially maps tile type identifiers (typedef'd as SHMACTileTypeID)
 * to tile instance constructor functions (typedef'd as SHMACTileConstructFunction)
 *
 * To implement auto-registration of tile types, simply implement a function
 *
 * void __registerTileTypes(SHMACLibrary * lib)
 * {
 *  lib->registerTileType('R', &constructRAMTile);
 *  lib->registerTileType('P', &constructPlasmaTile);
 *  lib->registerTileType('L', &constructLockTile);
 * }
 *
 * within any source file, and this function will be called upon SHMACLibrary
 * construction.
 */

#ifndef SHMACLIBRARY_H_
#define SHMACLIBRARY_H_

#include <SHMACTile.h>
#include <map>

// type for identifying a SHMAC tile
typedef  unsigned char SHMACTileTypeID;

// type info for a function used to construct a SHMACTile
typedef SHMACTile* (*SHMACTileConstructFunction)(shmacsim::tile_coord coord);

// lazy evaluated singleton pattern
class SHMACLibrary
{
public:
	static SHMACLibrary& getInstance()
	{
		static SHMACLibrary instance; // instantiated on first use, destruction guaranteed
		return instance;
	}


	bool registerTileType(SHMACTileTypeID identifier,
			SHMACTileConstructFunction constructFunction);
	bool unregisterTileType(SHMACTileTypeID identifier);

	SHMACTile * constructTile(SHMACTileTypeID identifier, shmacsim::tile_coord coord);

private:

	std::map<SHMACTileTypeID,SHMACTileConstructFunction> m_entries;

	SHMACLibrary(); // private constructor for singleton
	SHMACLibrary(SHMACLibrary const&); // do not implement
	void operator=(SHMACLibrary const&); // do not implement
};

#endif /* SHMACLIBRARY_H_ */
