/*
 * SHMACPacketTracker.cpp
 *
 *  Created on: Jun 19, 2013
 *      Author: maltanar
 */

#include "SHMACPacketTracker.h"

// weakly define the __tracePacket function - will only be called
// if it is defined regularly somewhere else
extern bool __tracePacket(shmacsim::packet *) __attribute__((weak));

SHMACPacketTracker::SHMACPacketTracker()
{
	m_entryChainHead = m_entryChainTail = NULL;
	m_trackingActive = false;
}

SHMACPacketTracker::~SHMACPacketTracker()
{
	if(m_trackingActive)
		terminateTracking();
}


void SHMACPacketTracker::tracePacket(shmacsim::packet * pkt, std::string loc)
{
	// only continue if __tracePacket is defined
	if(!__tracePacket)
		return;

	// call __tracePacket to determine whether this packet is to be tracked
	if(__tracePacket(pkt))
	{
		// configure tracker for this packet
		sc_assert(!(pkt->tag));
		SHMACPacketTracker * tracker = new SHMACPacketTracker();
		tracker->startTracking(pkt, loc);
		pkt->tag = (void*) tracker;
	}
}


void SHMACPacketTracker::startTracking(shmacsim::packet *pkt, std::string initialLocation)
{
	sc_assert(!m_trackingActive);
	sc_assert(!m_entryChainHead);

	m_entryChainHead = new PacketTrackingEntry;
	m_entryChainHead->start = sc_time_stamp();
	m_entryChainHead->locationName = initialLocation;
	m_entryChainHead->nextEntry = NULL;

	m_entryChainTail = m_entryChainHead;

	m_trackingActive = true;
	m_trackedPacket = *pkt;
}

void SHMACPacketTracker::terminateTracking()
{
	PacketTrackingEntry * entry = m_entryChainHead;
	PacketTrackingEntry * tmp;

	std::cerr << "*******************************************" << std::endl;
	std::cerr << "Tracking entries for " << m_trackedPacket << std::endl;

	sc_time eStart = m_entryChainHead->start;
	sc_time eEnd = (m_entryChainTail->end > m_entryChainTail->start ? m_entryChainTail->end : m_entryChainTail->start);

	while(entry != NULL)
	{
		if(entry->end > entry->start)
			std::cerr << "@" << entry->locationName << "\t\t\t" << entry->end-entry->start << "(" << entry->start << "..." << entry->end << ")" << std::endl;
		else
			std::cerr << "@" << entry->locationName << "\t\t\t(at " << entry->start << ")" << std::endl;

		tmp = entry;
		entry = entry->nextEntry;
		delete tmp;
	}

	std::cerr << "End-to-end: " << eEnd-eStart << " (" << eStart <<"..." << eEnd << ")" << std::endl;
	std::cerr << "*******************************************" << std::endl;
}

void SHMACPacketTracker::registerLocation(std::string locationName)
{
	sc_assert(m_trackingActive);

	if(m_entryChainTail->locationName == locationName)
	{
		// still in same location, tentatively set end
		m_entryChainTail->end = sc_time_stamp();
	} else
	{
		// finalize current entry
		m_entryChainTail->end = sc_time_stamp();
		m_entryChainTail->nextEntry = new PacketTrackingEntry;

		m_entryChainTail = m_entryChainTail->nextEntry;
		m_entryChainTail->start = sc_time_stamp();
		m_entryChainTail->locationName = locationName;
		m_entryChainTail->nextEntry = NULL;
	}
}
