/*
 * SHMACTile.cpp
 *
 *  Created on: 6 Mar 2013
 *      Author: Yaman Umuroglu
 */

#include "SHMACTile.h"
#include "SHMACDualPortRAM.h"
#include "shmac_simutils.h"
#include "shmac_typeutils.h"

SHMACTile::SHMACTile(shmacsim::tile_coord coords, SHMACRouter * router,
		SHMACMaster * master, SHMACSlave * slave) :
		SHMACElement(coords)
{
	// master and slave can be nonexistent, but the router is obligatory:
	sc_assert(router);

	m_router = router;
	m_master = master;
	m_slave = slave;
	// runtime initializer hook is not set by default
	m_runtimeInitializer = NULL;
	// bypass is not created by default
	m_bypass = NULL;

	initializeTileInternalConnections();
}

SHMACTile::~SHMACTile()
{
	if (m_bypass)
		delete m_bypass;
	if (m_router)
		delete m_router;
	if (m_master)
		delete m_master;
	if (m_slave)
		delete m_slave;
}

void SHMACTile::initializeTileInternalConnections()
{
	m_router->setParent(this);

	if (m_slave)
	{
		m_slave->setParent(this);
		m_router->connectSlave(m_slave);

		// handle dual-port slaves
		if (m_slave->getPortCount() > 1)
		{
			SHMACDualPortRAM * dualPortSlave = (SHMACDualPortRAM *) m_slave;
			char bypassName[15];
			sprintf(bypassName, "bp_%d_%d", (int) (getCoords().i),
					(int) (getCoords().j));
			// create bypass unit
			m_bypass = new SHMACRouterBypass(bypassName);
			// set local memory range
			SHMAC_ADDR memStart = setAddressTileCoord(0, getCoords());
			m_bypass->setLocalMemRange(memStart, memStart + m_slave->getSize());
			// connect dual-port slave to bypass and vice versa
			m_bypass->toRAMPort.bind(dualPortSlave->secondaryInboundPort);
			dualPortSlave->secondaryOutboundPort.bind(*m_bypass);
		}
	}

	if (m_master)
	{
		m_master->setParent(this);
		if(!m_bypass)
			m_router->connectMaster(m_master);	// connect master directly to router
		else
		{
			// connect router <--> bypass
			m_router->connectSHMACMemInterface(shmacsim::DIR_INTERN_PROC, m_bypass->fromRouterPort, m_bypass->toRouterPort);
			// connect bypass <--> master
			m_bypass->toMasterPort.bind(m_master->inboundMemoryPort);
			m_master->outboundMemoryPort.bind(*m_bypass);
		}
	}
}

SHMACMaster* SHMACTile::getMaster() const
{
	return m_master;
}

SHMACRouter* SHMACTile::getRouter() const
{
	return m_router;
}

SHMACSlave* SHMACTile::getSlave() const
{
	return m_slave;
}

void SHMACTile::setRuntimeInitializer(SHMACTileRuntimeInitializer func)
{
	m_runtimeInitializer = func;
}

void SHMACTile::executeRuntimeInitializer()
{
	// execute the runtime initializer hook, if set
	if (m_runtimeInitializer)
		(*m_runtimeInitializer)(this);
}

void SHMACTile::setMetaDataString(std::string key, std::string value)
{
	m_metaDataStr[key] = value;
}

std::string SHMACTile::getMetaDataString(std::string key)
{
	return m_metaDataStr[key];
}

void SHMACTile::setMetaDataInt(std::string key, unsigned int value)
{
	m_metaDataInt[key] = value;
}

unsigned int SHMACTile::getMetaDataInt(std::string key)
{
	return m_metaDataInt[key];
}

void SHMACTile::connectToTile(SHMACTile * dest)
{
	// ensure destination tile exists
	sc_assert(dest);

	// deduce neighborhood relationship based on coords
	shmacsim::tile_coord c1 = this->getCoords();
	shmacsim::tile_coord c2 = dest->getCoords();
	int di = c1.i - c2.i, dj = c1.j - c2.j;

	// this connectTo handles horizontal or vertical neighbours
	if (abs(di) + abs(dj) != 1)
	{
		std::cerr << "SHMACTile::connectTo used on non-neighbouring tiles!"
				<< std::endl;
		sc_assert(0);
	}

	SHMACRouter * router1 = this->getRouter();
	SHMACRouter * router2 = dest->getRouter();

	// connect routers
	// shmacsim::directionOfCoords is used for dimension order
	// routing, but since we are immediate neighbours with dest it
	// works fine here to deduce the neighbor's direction
	router1->connectToRouter(router2, shmacsim::directionOfCoords(c1, c2));
}
