// tile_types.cpp
// source file for defining tile types used in SHMACsim
// these tile types will be registed with SHMACLibrary and
// can be later instantiated by their corresponding
// identifiers

#ifndef TILE_TYPES_H_
#define TILE_TYPES_H_

#include "SHMACLibrary.h"

#include "SHMACTile.h"
#include "SHMACRouter.h"
#include "SHMACSlave.h"
#include "SHMACMasterArchCMIPS1.h"
#include "SHMACLockSlave.h"
#include "SHMACRouterCycAcc.h"
#include "SHMACJumpSlave.h"
#include "SHMACClockSlave.h"
#include "SHMACDualPortRAM.h"
#include "shmac_simutils.h"

// whether to use the approximately timed or the cycle accurate router model
#define ROUTER_MODEL_CA

// constructor functions for each tile type
SHMACTile * constructRAMTile(shmacsim::tile_coord coord);
SHMACTile * constructPlasmaTile(shmacsim::tile_coord coord);
SHMACTile * constructLockTile(shmacsim::tile_coord coord);
SHMACTile * constructDualPortMemPlasmaTile(shmacsim::tile_coord coord);
SHMACTile * constructJumpTile(shmacsim::tile_coord coord);
SHMACTile * constructClockTile(shmacsim::tile_coord coord);

// runtime initializer functions for the tiles that need it
void initializePlasmaTile(SHMACTile * tile);

// defines used by tile types, these could be metadata parameter names
// or anything else that is required
#define		PLASMATILE_PARAM_LOAD_BIN				"sLoadBinary"
#define		PLASMATILE_PARAM_CMDLINE				"sCmdLine"
#define		PLASMATILE_PARAM_HEAP_PTR				"iHeapPtr"
#define		PLASMATILE_PARAM_ENABLE_INTG_BOOTSTRAP	"iIntgBoot"

// clock signal, for any modules which may require it
sc_clock * global_clk = NULL;


// registers the tile types provided in this file with SHMACLibrary
void __registerTileTypes(SHMACLibrary * lib)
{
	lib->registerTileType('R', &constructRAMTile);
	lib->registerTileType('P', &constructPlasmaTile);
	lib->registerTileType('L', &constructLockTile);
	lib->registerTileType('D', &constructDualPortMemPlasmaTile);
	lib->registerTileType('J', &constructJumpTile);
	lib->registerTileType('C', &constructClockTile);

	// FIXME create the clk in a more appropriate place!
	global_clk = new sc_clock("clk", SHMAC_CLOCK_PERIOD_NS, SC_NS, 0.5);

#ifdef ROUTER_MODEL_CA
	std::cerr << "Using cycle-accurate router model" << std::endl;
#else
	std::cerr << "Using approximate-timed router model" << std::endl;
#endif
}


SHMACTile * constructJumpTile(shmacsim::tile_coord coord)
{
	static char nm[10];

	sprintf(nm, "r_%d_%d", coord.i, coord.j);
#ifndef ROUTER_MODEL_CA
	SHMACRouter * router = new SHMACRouter(nm, NULL);
#else
	SHMACRouter * router = new SHMACRouterCycAcc(nm, NULL);
	((SHMACRouterCycAcc*) router)->clk(*global_clk);
#endif
	sprintf(nm, "js_%d_%d", coord.i, coord.j);
	SHMACSlave * slave = (SHMACSlave *) new SHMACJumpSlave(nm, NULL);

	return new SHMACTile(coord, router, NULL, slave);
}

SHMACTile * constructClockTile(shmacsim::tile_coord coord)
{
	static char nm[10];

	sprintf(nm, "r_%d_%d", coord.i, coord.j);
#ifndef ROUTER_MODEL_CA
	SHMACRouter * router = new SHMACRouter(nm, NULL);
#else
	SHMACRouter * router = new SHMACRouterCycAcc(nm, NULL);
	((SHMACRouterCycAcc*) router)->clk(*global_clk);
#endif
	sprintf(nm, "cs_%d_%d", coord.i, coord.j);
	SHMACSlave * slave = (SHMACSlave *) new SHMACClockSlave(nm, NULL);

	return new SHMACTile(coord, router, NULL, slave);
}

SHMACTile * constructRAMTile(shmacsim::tile_coord coord)
{
	static char nm[10];

	sprintf(nm, "r_%d_%d", coord.i, coord.j);
#ifndef ROUTER_MODEL_CA
	SHMACRouter * router = new SHMACRouter(nm, NULL);
#else
	SHMACRouter * router = new SHMACRouterCycAcc(nm, NULL);
	((SHMACRouterCycAcc*) router)->clk(*global_clk);
#endif
	sprintf(nm, "s_%d_%d", coord.i, coord.j);
	SHMACSlave * slave = new SHMACSlave(nm, 16 * 1024 * 1024, NULL);

	return new SHMACTile(coord, router, NULL, slave);
}

SHMACTile * constructLockTile(shmacsim::tile_coord coord)
{
	static char nm[10];

	sprintf(nm, "r_%d_%d", coord.i, coord.j);
#ifndef ROUTER_MODEL_CA
	SHMACRouter * router = new SHMACRouter(nm, NULL);
#else
	SHMACRouter * router = new SHMACRouterCycAcc(nm, NULL);
	((SHMACRouterCycAcc*)router)->clk(*global_clk);
#endif
	sprintf(nm, "lt_%d_%d", coord.i, coord.j);
	SHMACSlave * slave = new SHMACLockSlave(nm, 16 * 1024 * 1024, NULL);

	return new SHMACTile(coord, router, NULL, slave);
}

void initializePlasmaTile(SHMACTile * tile)
{
	unsigned int importSize = 0;
	std::string binPath = tile->getMetaDataString(PLASMATILE_PARAM_LOAD_BIN);
	if (!binPath.empty())
	{
		std::cerr << "Loading " << binPath << " onto " << tile->getCoords()
				<< std::endl;
		tile->getSlave()->importContents(binPath.c_str(), &importSize);
	}

	// need to set the heap pointer in the ArchC MIPS core for
	// correct functioning of any functions that use the heap
	// use the specified value, if any
	// otherwise, use the import size
	if (tile->getMetaDataInt(PLASMATILE_PARAM_HEAP_PTR))
		importSize = tile->getMetaDataInt(PLASMATILE_PARAM_HEAP_PTR);
	std::cerr << "Initializing heap pointer for ArchC MIPS core at "
			<< tile->getCoords() << " to " << importSize << std::endl;

	// enable/disable integrated bootstrapping
	// (to get rid of jump tile)
	bool integratedBootstrap = false;
	if (tile->getMetaDataInt(PLASMATILE_PARAM_ENABLE_INTG_BOOTSTRAP) == 1)
		integratedBootstrap = true;

	// get any desired command line args to be passed to this app
	std::string cmdLine = tile->getMetaDataString(PLASMATILE_PARAM_CMDLINE);


	((SHMACMaster_ArchC_MIPS1 *) (tile->getMaster()))->prepare(importSize, cmdLine, integratedBootstrap);
}

SHMACTile * constructPlasmaTile(shmacsim::tile_coord coord)
{
	static char nm[10];

	sprintf(nm, "r_%d_%d", coord.i, coord.j);
#ifndef ROUTER_MODEL_CA
	SHMACRouter * router = new SHMACRouter(nm, NULL);
#else
	SHMACRouter * router = new SHMACRouterCycAcc(nm, NULL);
	((SHMACRouterCycAcc*)router)->clk(*global_clk);
#endif
	sprintf(nm, "c_%d_%d", coord.i, coord.j);
	SHMACMaster_ArchC_MIPS1 * core = new SHMACMaster_ArchC_MIPS1(nm, NULL);
	sprintf(nm, "s_%d_%d", coord.i, coord.j);
	SHMACSlave * slave = new SHMACSlave(nm, 16 * 1024 * 1024, NULL);

	SHMACTile * tile = new SHMACTile(coord, router, core, slave);

	// set the runtime initializer for the tile
	tile->setRuntimeInitializer(&initializePlasmaTile);

	return tile;
}

SHMACTile * constructDualPortMemPlasmaTile(shmacsim::tile_coord coord)
{
	static char nm[15];

	sprintf(nm, "r_%d_%d", coord.i, coord.j);
#ifndef ROUTER_MODEL_CA
	SHMACRouter * router = new SHMACRouter(nm, NULL);
#else
	SHMACRouter * router = new SHMACRouterCycAcc(nm, NULL);
	((SHMACRouterCycAcc*)router)->clk(*global_clk);
#endif
	sprintf(nm, "c_%d_%d", coord.i, coord.j);
	SHMACMaster_ArchC_MIPS1 * core = new SHMACMaster_ArchC_MIPS1(nm, NULL);
	sprintf(nm, "dps_%d_%d", coord.i, coord.j);
	SHMACSlave * slave = (SHMACSlave *) new SHMACDualPortRAM(nm, 16 * 1024 * 1024, NULL);

	SHMACTile * tile = new SHMACTile(coord, router, core, slave);

	// set the runtime initializer for the tile
	tile->setRuntimeInitializer(&initializePlasmaTile);

	return tile;
}

#endif /* TILE_TYPES_H_ */
