#include "shmac_simutils.h"

sc_time shmacsim::getClockCycles(unsigned int n)
{
	sc_time ret(SHMAC_CLOCK_PERIOD_NS, SC_NS);

	return n * ret;
}

void shmacsim::waitClockCycles(unsigned int n)
{
	wait(getClockCycles(n));
}

SHMACGrid * shmacsim::createGridFromConfigFile(const char * fileName)
{
	ifstream file;
	char config[SHMAC_MAX_GRIDDIM][SHMAC_MAX_GRIDDIM];
	file.open(fileName, ios::in);
	unsigned int row_count = 0, col_count = 0;
	unsigned int line_len = 0;

	if (!file.is_open())
	{
		std::cerr << "createGridFromConfigFile: could not open " << fileName
				<< std::endl;
		return NULL;
	}

	while (file.good())
	{
		file.getline(config[row_count], SHMAC_MAX_GRIDDIM);
		line_len = strlen(config[row_count]);
		// ensure all columns are of the same length
		// i.e we want a grid, not a triangle :)
		if (!col_count)
			col_count = line_len;
		else if (!line_len) // empty liine, assume end of file
			break;
		else if (col_count != line_len)
		{
			std::cerr << "createGridFromConfigFile: column size mismatch"
					<< std::endl;
			file.close();
			return NULL;
		}
		std::cerr << config[row_count] << std::endl;
		row_count++;
	}

	file.close();
	std::cerr << "Creating SHMACGrid with size " << row_count << "x"
			<< col_count << " (rowsxcols)" << std::endl;
	SHMACGrid * theGrid = new SHMACGrid(row_count, col_count);

	// construct grid with tile types given in config file
	for (uint8_t i = 0; i < row_count; i++)
		for (uint8_t j = 0; j < col_count; j++)
			(*theGrid)(i, j) = SHMACLibrary::getInstance().constructTile(
					config[i][j], { i, j });

	// connect the neighbouring tiles
	theGrid->establishConnections();

	return theGrid;
}

void shmacsim::configureGridFromConfigFile(const char * configFile,
		SHMACGrid * gr)
{
	sc_assert(gr);

	int i, j;
	unsigned int rows, cols;
	char val[1024], keyName[10];

	rows = gr->getRowCount();
	cols = gr->getColCount();

	FILE * runtimeConfig = fopen(configFile, "r");
	while (!feof(runtimeConfig))
	{
		// read a line from the configuration file
		// the format is expected to look like (%d,%d)\t[key]\t[value], so for example
		// (0,1)	/home/maltanar/binary_file.bin
		// use %[^\t\n] to read in the key and value to support non-tab whitespace there
		fscanf(runtimeConfig, "(%d,%d)\t%[^\t\n]\t%[^\t\n]\n", &i, &j, keyName,
				val);
		// check that the coordinates are valid
		if (i == -1 && j == -1)
		{
			// special case: (-1, -1) is treated as the entire grid
			// no special action needed, handled in setter part below
		}
		else if (i < 0 || i >= rows)
		{
			std::cerr << "configureGridFromConfigFile: i coordinate " << i
					<< " is invalid (iMax " << rows << ") -- skipping"
					<< std::endl;
			continue;
			//fclose(runtimeConfig);
			//return;
		}
		else if (j < 0 || j >= cols)
		{
			std::cerr << "configureGridFromConfigFile: j coordinate " << j
					<< " is invalid (jMax " << cols << ") -- skipping"
					<< std::endl;
			continue;
			//fclose(runtimeConfig);
			//return;
		}

		if (i == -1 && j == -1)
			std::cerr << "Setting " << keyName << "=" << val << " on all tiles"
					<< std::endl;
		else
			std::cerr << "Setting " << keyName << "=" << val << " on (" << i
					<< "," << j << ")" << std::endl;
		// set metadata on desired tile(s), according to type prefix:
		// s : string
		// i : integer
		std::string str_key(keyName), str_val(val);
		unsigned int int_val;
		if (keyName[0] == 'i')
		{
			sscanf(val, "%u", &int_val);
			if (i == -1 && j == -1)
				gr->setMetaDataInt(str_key, int_val);
			else
				(*gr)(i, j)->setMetaDataInt(str_key, int_val);
		}
		else if (keyName[0] == 's')
		{
			if (i == -1 && j == -1)
				gr->setMetaDataString(str_key, str_val);
			else
				(*gr)(i, j)->setMetaDataString(str_key, str_val);
		}
	}

	fclose(runtimeConfig);
}

shmacsim::direction_enum shmacsim::directionOfCoords(
		shmacsim::tile_coord localCoords, shmacsim::tile_coord foreignCoords)
{

	// dimension order routing:
	// first, route along i axis until in correct j position

	int dj = foreignCoords.j - localCoords.j;
	int di = foreignCoords.i - localCoords.i;

	if (dj > 0)
		return shmacsim::DIR_EAST;
	else if (dj < 0)
		return shmacsim::DIR_WEST;
	else if (di > 0)
		return shmacsim::DIR_SOUTH;
	else if (di < 0)
		return shmacsim::DIR_NORTH;
	else
		return shmacsim::DIR_INTERN_MEM;
}

int32_t shmacsim::convertHostToTargetEndian_32bit(int32_t val)
{
	// FIXME need a way to configure host-target endianness relation
	return __builtin_bswap32(val);
}

uint32_t shmacsim::procIDFromCoords(shmacsim::tile_coord coord)
{
	return (coord.i * SHMAC_MAX_GRIDDIM) + coord.j;
}

shmacsim::tile_coord shmacsim::coordsFromProcID(uint32_t procID)
{
	shmacsim::tile_coord crd;
	crd.i = procID / SHMAC_MAX_GRIDDIM;
	crd.j = procID % SHMAC_MAX_GRIDDIM;

	return crd;
}
