/*
 * shmac_types.h
 *
 * SHMAC simulator internal typedefs and structures
 * adapted from package.vhd from SHMAC 2012 MSc thesis
 *
 *  Created on: 21 Feb 2013
 *      Author: Yaman Umuroglu
 */

#ifndef SHMAC_TYPES_H_
#define SHMAC_TYPES_H_

#include <systemc.h>

#define		SHMAC_WORD		shmacsim::bus32
#define		SHMAC_ADDR		shmacsim::bus32

namespace shmacsim
{

typedef uint32_t bus32;
typedef uint8_t leds_t;

// address with all zeroes - used for booting each processor
// accessing this address jumps to start of local tile address space
#define		SHMAC_BOOT_ADDRESS						0x00000000

// size of a single coordinate in bits
#define		SHMAC_BSOF_COORD						4

// maximum possible value for a coordinate
#define		SHMAC_MAX_COORD							((1 << SHMAC_BSOF_COORD) - 1)

// maximum possible value for a grid dimension
#define		SHMAC_MAX_GRIDDIM						(1 << SHMAC_BSOF_COORD)

// type corresponding to a single coordinate
typedef uint8_t coord_int;

// number of coordinates
#define		SHMAC_COORD_COUNT						2

// location of a tile on the grid
typedef struct tile_coord
{
	coord_int i;
	coord_int j;

	tile_coord& operator=(const tile_coord &rhs);
	bool operator==(const tile_coord & rhs) const;
} tile_coord;

// SHMAC mem address looks like "iiiijjjjaaaaaaaaaaaaaaaaaaaaaaaa" (bitwise)
// where i and j are tile coords, a is the local address inside that tile
// the define's below are relevant for extracting coords from the address
#define		SHMAC_BSOF_ADDR							32
#define		SHMAC_J_SHIFT							24
#define		SHMAC_I_SHIFT							(SHMAC_J_SHIFT + SHMAC_BSOF_COORD)
#define		SHMAC_COORD_MASK						(bus32)((1 << SHMAC_BSOF_COORD) - 1)
#define		SHMAC_LADDR_MASK						(bus32)((1 << SHMAC_J_SHIFT) - 1)

// directions
typedef enum
{
	DIR_EAST, DIR_NORTH, DIR_WEST, DIR_SOUTH, DIR_INTERN_PROC, DIR_INTERN_MEM
} direction_enum;

// number of entries in direction_enum
#define		SHMAC_DIRECTION_COUNT					6

// useful types derived from direction enum
typedef direction_enum direction_enum_vector[SHMAC_DIRECTION_COUNT];
typedef sc_bv<SHMAC_DIRECTION_COUNT> direction_bool_vector;
typedef sc_lv<SHMAC_DIRECTION_COUNT> direction_stdlogic_vector;

// router package flags
typedef enum
{
	PACKET_FLAG_NONE, // no flags
	PACKET_FLAG_INTR, // interrupt flag
	PACKET_FLAG_SYNCO // sync flag
} packet_flag;

// possible values for the packet write_enable field
typedef enum
{
	PACKET_WREN_READ4, // read 4 bytes
	PACKET_WREN_WRITE1_P1, // write 1 byte, bits 0..7
	PACKET_WREN_WRITE1_P2, // write 1 byte, bits 8..15
	PACKET_WREN_WRITE1_P3, // write 1 byte, bits 16..23
	PACKET_WREN_WRITE1_P4, // write 1 byte, bits 24..31
	PACKET_WREN_WRITE2_P1, // write 2 bytes, bits 0..15
	PACKET_WREN_WRITE2_P2, // write 2 bytes, bits 16..31
	PACKET_WREN_WRITE4 // write 4 bytes
} packet_wren;

#define		PACKET_WREN_LEN					4

//SHMAC router packet structure
typedef struct packet
{
	bool req; // set if packet is a request, not set if response
	packet_wren write_enable; // read 4 bytes, or write 1/2/4 bytes
	packet_flag flag; // enumeration for flag values
	tile_coord sender; // issuer of request (even if this is a corresponding response)
	bus32 address; // location of requested data (even if this is a corresponding response)
	bus32 data; // data, always 4 bytes, content ignored for reads

	void * tag;	// SHMACsim only - for attaching tags or tracking objects

	// need to define assignment and comparison operators
	// to be able to use this type over SystemC channels
	packet& operator=(const packet& rhs);
	bool operator==(const packet& rhs) const;

	inline friend void sc_trace(sc_trace_file* tf, const shmacsim::packet& val,
			const std::string& nm)
	{
		sc_trace(tf, val.address, nm + ".addr");
		sc_trace(tf, val.data, nm + ".data");
	}
	;

	inline friend ostream& operator<<(ostream& file, const packet& pkt)
	{
		return file << "SHMAC packet {" << std::endl << "req = " << pkt.req
				<< " addr = " << std::hex << pkt.address << std::dec << ", data = "
				<< pkt.data << ", r/w = " << pkt.write_enable << ", flags = " << pkt.flag
				<< ", sender = (" << (int)(pkt.sender.i) << ", " << (int)(pkt.sender.j) << ")" << std::endl
				<< "}" << std::endl;
	}
	;

} packet;

// vector of packets indexed by direction
typedef packet packet_vector[SHMAC_DIRECTION_COUNT];

}

ostream& operator<<(ostream& file, shmacsim::tile_coord const & dbv);

#endif /* SHMAC_TYPES_H_ */
