#include "SHMACSysRegFile.h"
#include "shmac_simutils.h"
#include "SHMACJumpSlave.h"

SHMACSysRegFile::SHMACSysRegFile(sc_module_name name) :
		sc_module(name)
{
	m_enableBootstrap = false;
}

void SHMACSysRegFile::receiveMessage(shmacsim::packet message)
{
	// intercept packets from processor dealing with sysreq registers
	// or bootstrap
	if (message.req && message.address >= SHMAC_SYSREG_MIN
			&& message.address <= SHMAC_SYSREG_MAX)
		handleSysRegReqPacket(message);
	else if (isBootstrap(message))
		handleSysRegReqPacket(message);
	else
		handleRegularPacket(message);

}

void SHMACSysRegFile::handleSysRegReqPacket(shmacsim::packet message)
{
	shmacsim::packet response = message;
	shmacsim::tile_coord requesterCoord = message.sender;
	response.req = false;

	switch (message.address)
	{
	case SHMAC_SYSREG_CLK:
		// clock register
		// replaces clock tile functionality
		// return elapsed CCs since start of simulation
		response.data = (sc_time_stamp() / shmacsim::getClockCycles(1));
		break;
	case SHMAC_SYSREG_TILEID:
		// tile ID register
		// returns tile ID (generated based on coords)
		// can be used for bootstrapping
		response.data = getTileID();
		break;
		// FIXME operations defined below are MIPS-specific
	case SHMAC_JUMPSLAVE_ADDR_LUI:
		//  lui         0     rt       i     j       0
		// "001111"&"00000"&"01100"&"iiii"&"jjjj"&"00000000"
		// FIXME masks should be dependent on coordinate size
		response.data = 0x3C0C0000 | ((requesterCoord.i & 0x0F) << 12)
				| ((requesterCoord.j & 0x0F) << 8);
		break;
	case SHMAC_JUMPSLAVE_ADDR_ORI:
		//    ori       rs     rt      imm
		// "001101"&"01100"&"01100"&"0000000000000000"
		// OR register 12 with itself
		response.data = 0x358C0000;
		break;
	case SHMAC_JUMPSLAVE_ADDR_JR:
		//   special   rs          0       hint     jr
		// "000000"&"01100"&"0000000000"&"00000"&"001000"
		response.data = 0x01800008;
		break;
	case SHMAC_JUMPSLAVE_ADDR_NOP:
		// nop, all zeroes
		response.data = 0x0;
		break;
	default:
		sc_assert(0);
		// undefined sysreg
		break;
	}

	// std::cerr << "SHMACSysRegFile: returning clock register " << response.data << std::endl;

	// this operation crosses the host-target mem boundary, do endian conversion
	response.data = shmacsim::convertHostToTargetEndian_32bit(response.data);

	// send response to core
	cpuPort->receiveMessage(response);
}

bool SHMACSysRegFile::isBootstrap(shmacsim::packet message)
{
	// check whether bootstrapping is enabled first
	if (!m_enableBootstrap)
		return false;

	// return true if this packet is within bootstrap instr range
	return (message.address >= SHMAC_JUMPSLAVE_ADDR_LUI
			&& message.address <= SHMAC_JUMPSLAVE_ADDR_NOP);
}

void SHMACSysRegFile::handleRegularPacket(shmacsim::packet message)
{
	// forward packet to either CPU or router depending on request flag
	if (message.req)
		// forwarded to router
		routerPort->receiveMessage(message);
	else
		// forwarded to core
		cpuPort->receiveMessage(message);
}
