/*
 * SHMACTLMAdapter.cpp
 *
 * adapter for converting between ArchC TLM memory IF <-> SHMAC memory IF
 *
 *  Created on: 4 Mar 2013
 *      Author: Yaman Umuroglu
 */

#include "SHMACTLMAdapter.h"
#include "shmac_typeutils.h"
#include "shmac_simutils.h"
#include "SHMACPacketTracker.h"

SHMACTLMAdapter::SHMACTLMAdapter(sc_module_name name, SHMACElement * parent) :
		m_parent(parent), sc_module(name)
{
	m_scExecuted = m_locked = false;
	m_lockAddress = m_lockReturnValue = 0;
	m_pendingRequest = false;
}

// transfer a ArchC TLM request into a SHMAC req pkt and send it
ac_tlm_rsp SHMACTLMAdapter::transport(const ac_tlm_req& tlm_request)
{
	// cannot function without parent
	sc_assert(m_parent);

	ac_tlm_rsp tlm_response;

	tlm_response.req_type = tlm_request.type;
	tlm_response.status = SUCCESS;

	// create SHMAC request packet with given destination
	shmacsim::packet shmacmem_request = makeRequestPacket(m_parent->getCoords(),
			tlm_request.addr);

	shmacsim::packet shmacmem_response;

	// need to handle translation between ArchC TLM <-> SHMAC mem packets
	// LL/SC is based on SC success (lockless) while ArchC TLM only defines
	// LOCK and UNLOCK operations. we change the semantics of LOCK/UNLOCK as:
	// - LL call is made as LOCK, READ, UNLOCK, works just like regular load
	// - SC call is LOCK, WRITE, READ, UNLOCK. the READ returns the success
	//   of the preceding WRITE instruction. this is needed since ArchC does
	//   not examine the return value of the WRITE request.
	// - regular READ and WRITE (without locking calls) generate normal r/w
	// FIXME this workaround is not the most elegant way. change ArchC TLM IF?
	switch (tlm_request.type)
	{
	case LOCK:
		sc_assert(!m_locked);
		m_locked = true;
		m_scExecuted = false;
		break;
	case UNLOCK:
		sc_assert(m_locked);
		m_locked = false;
		break;
	case READ:
		if (!m_locked)
		{
			// no locking, regular load from memory
			shmacmem_request.write_enable = shmacsim::PACKET_WREN_READ4;
			shmacmem_response = sendSHMACMessage(shmacmem_request);

			// interpret SHMAC response packet and convert to TLM response
			sc_assert(shmacmem_response.req == false);
			tlm_response.data = shmacmem_response.data;
		}
		else
		{
			// special case: reading while "locked"
			if (m_scExecuted)
			{
				// case 1: SC was already executed
				// performs no actual memory operations
				// returns whether SC was successful
				sc_assert(tlm_request.addr == m_lockAddress);
				tlm_response.data = m_lockReturnValue;
			}
			else
			{
				// case 2: SC not executed
				// this is an LL instruction
				shmacmem_request.write_enable = shmacsim::PACKET_WREN_READ4;
				shmacmem_request.flag = shmacsim::PACKET_FLAG_SYNCO;
				shmacmem_response = sendSHMACMessage(shmacmem_request);

				// interpret SHMAC response packet and convert to TLM response
				sc_assert(shmacmem_response.req == false);
				tlm_response.data = shmacmem_response.data;
			}
		}
		break;
	case WRITE:
		// copy data into SHMAC request packet
		shmacmem_request.data = tlm_request.data;
		shmacmem_request.write_enable = shmacsim::PACKET_WREN_WRITE4;
		// add sync flag wrt lock status (differs SC from normal store)
		shmacmem_request.flag =
				(m_locked ?
						shmacsim::PACKET_FLAG_SYNCO : shmacsim::PACKET_FLAG_NONE);
		shmacmem_response = sendSHMACMessage(shmacmem_request);

		// special case: writing while locked
		// here we actually stall for a response and make use of the
		// returned response's sync flag to determine SC success
		if (m_locked)
		{
			sc_assert(shmacmem_response.req == false);

			// store SC'd address (for assertion while locked reading)
			m_lockAddress = tlm_request.addr;
			// SC has now been executed, set flag to true
			// allows distinguishing the locked load in LL and locked load in SC
			m_scExecuted = true;
			// check sync flag, which indicates the SC success
			if (shmacmem_response.flag == shmacsim::PACKET_FLAG_SYNCO)
				m_lockReturnValue = 1;
			else
				m_lockReturnValue = 0;
		} else
			// for other writes, the return value should be ignored (since SHMAC
			// does not even stall and wait for a response).
			// if mysterious 0xdeadbeef's start turning up during program behaviour,
			// someone is making use of the write return data, which is WROOONG
			tlm_response.data = 0xdeadbeef;
		break;
	default:
		sc_assert(0);
		// unrecognized TLM request type
		break;
	}

	return tlm_response;
}

// called by the router when a message is destined for this adapter
// (or rather, the connected processor)
void SHMACTLMAdapter::receiveMessage(shmacsim::packet message)
{
	// fromSrc can be ignored since the router is the sole source of packets for the processor
	// however, if we use a dual-memory-port with the local slave, this may have to change!
	// TODO add assertions: coordinate correctness
	// make sure that we are expecting a reply, and the reply is
	// for the address that we are expecting
	sc_assert(m_pendingRequest);
	sc_assert(m_requestAddress == message.address);
	// make an internal copy of the received reply
	m_receivedData = message;

	if(message.tag)
	{
		((SHMACPacketTracker *)message.tag)->registerLocation("TLMAdapter");
	}

	//std::cerr << "TLM recv\t" << sc_time_stamp() << "\t" << std::hex << message.address << "\t" << message.data << std::dec << std::endl;

	// notify the sync/async border that we have received the reply
	// in the non-CA router model, this will be executed in the same context as wait(),
	// so we need a delayed notify for this to work correctly
	// zero time makes no difference for the CA model either, so use this for both
	m_replyReceived.notify(SC_ZERO_TIME);
}

// declared virtual, used for the sync/async boundary
// also simulates the effects of the plasma wrapper (memory stall state machine)
shmacsim::packet SHMACTLMAdapter::sendSHMACMessage(shmacsim::packet packet)
{
	// starts packet tracking, if desired
	SHMACPacketTracker::tracePacket(&packet, "TLMAdapter");

	// indicate that we are waiting for a reply, and for which address
	m_pendingRequest = true;
	m_requestAddress = packet.address;

	//std::cerr << "TLM send\t" << sc_time_stamp() << "\t" << std::hex << packet.address << std::dec << "\t" << packet.write_enable << std::endl;

	shmacsim::waitClockCycles(1);

	// send out the memory request
	shmacMemPort->receiveMessage(packet);

	//std::cerr << "TLM send OK\t" << sc_time_stamp() << "\t" << std::hex << packet.address << std::dec << "\t" << packet.write_enable << std::endl;

	// memory state machine delays 1 cycle here
	// FIXME this is causing freezes in the non-CA version, as well as
	// the sysreg accesses on the CA version
	//shmacsim::waitClockCycles(1);

	// read requests require a reply, stalling the processor until
	// the reply is received, whereas write requests have no stalling
	// the LL/SC operations (indicated by the presence of the sync flag)
	// also stall until reply is received
	if (packet.write_enable == shmacsim::PACKET_WREN_READ4
			|| packet.flag == shmacsim::PACKET_FLAG_SYNCO)
	{
		wait(m_replyReceived);
		// memory state machine delays
		shmacsim::waitClockCycles(1);
	}

	// memory state machine delays
	shmacsim::waitClockCycles(1);

	// terminate tracking, if tracking info exists
	if (m_receivedData.tag)
	{
		((SHMACPacketTracker *) m_receivedData.tag)->registerLocation("TLMAdapter");
		delete ((SHMACPacketTracker *) m_receivedData.tag);
		m_receivedData.tag = NULL;
	}

	// clear pending request flags
	m_pendingRequest = false;

	return m_receivedData;
}
