/*
 * SHMACRouter.cpp
 *
 *  Created on: 26 Feb 2013
 *      Author: Yaman Umuroglu
 */

#include "SHMACRouter.h"
#include "shmac_typeutils.h"
#include "shmac_simutils.h"

SHMACRouter::SHMACRouter(sc_module_name instanceName, SHMACElement * parent) :
		sc_module(instanceName), m_parent(parent)
{

}

// for the moment, just handles sending a request as a call chain to appropriate
// sendMessage functions in appropriate elements & returns the response packet
void SHMACRouter::receiveMessage(shmacsim::packet message)
{
	// need to have set a parent to be able to function
	sc_assert(m_parent);
	// TODO remove this if it's completely OK
	// must be a request for us to be able to handle it
	// sc_assert(message.req == true);

	// special case: detect accesses to global address 0
	// bootstrapping the local processor, jump to local address start
	/*if (message.address == SHMAC_BOOT_ADDRESS)
	 return handleBootstrapping(message);*/

	// get packet destination
	shmacsim::direction_enum destDir;

	if (message.req)
		// determine who handles this request next
		destDir = m_parent->directionOfCoords(
				getTileCoordFromAddress(message.address));
	else
	{
		// if this is a reply packet for this tile...
		if (message.sender == m_parent->getCoords())
			// special case: override routing destination to INTERN_PROC
			destDir = shmacsim::DIR_INTERN_PROC;
		else
			// replies are sent to the request originator
			destDir = m_parent->directionOfCoords(message.sender);
	}

	// FIXME make router debug trace optional - remove
	/*if(message.address == 4)
	 std::cerr << sc_time_stamp() << ": router at " << m_parent->getCoords() << " routing pkt "<< message << " to "
	 << getDirectionName(destDir) << std::endl;*/
	// FIXME temp fix to get non-CA model working
	shmacsim::waitClockCycles(13);

	// FIXME also need forwarding to internal processor if we eventually decide to use actual 2 channels
	//sc_assert(destDir != shmacsim::DIR_INTERN_PROC);
	//sc_assert(destDir == shmacsim::DIR_INTERN_MEM); // FIXME remove to test multiproc

	// check if we can actually forward to this destination
	if (!routerPorts[destDir].get_interface())
	{
		std::cerr << "SHMACRouter::sendMessage: destination is invalid!"
				<< " router at  " << m_parent->getCoords() << " message dest: "
				<< std::hex << message.address << std::dec << " to direction: "
				<< getDirectionName(destDir)<< std::endl;
				sc_assert(0);
			}

		// a way to model the concurrency inside the router is to view the output port as a shared resource
		// here, we lock a mutex corresponding to the output port and add a small timed wait before
		// making the call to the next receiveMessage
		// since the entire set of output ports along the route have to be free simultaneously in order for
		// the call to be completed, this doesn't very accurately reflect SHMAC. this is more like circuit
		// switching, whereas the SHMAC network is packet switched. but it is an interesting experiment
		// anyway.
	m_lock[destDir].lock();
	wait(1, SC_NS);
	routerPorts[destDir]->receiveMessage(message);
	m_lock[destDir].unlock();

	// forward message to determined destination
	// calculate message origin direction by getting the inverse of dest dir
	// note that this is irrelevant for the CPU and mem receiveMessage implementations,
	// since they only receive messages from the router anyway

}

void SHMACRouter::connectToRouter(SHMACRouter *dest,
		shmacsim::direction_enum directionTowardsDest)
{
	sc_assert(dest);

	// ensure router models are compatible - cannot e.g connect
	// RTL router model to untimed router model
	if (dest->getRouterModelType() != this->getRouterModelType())
	{
		std::cerr << "Router models are not compatible: " << std::endl
				<< this->getParent()->getCoords() << " has "
				<< this->getRouterModelType() << std::endl
				<< dest->getParent()->getCoords() << " has "
				<< dest->getRouterModelType() << std::endl;
		sc_assert(0);
	}

	// model-specific section: establish connects towards dest

	// ensure this direction has not been connected before
	sc_assert(routerPorts[directionTowardsDest].bind_count() == 0);
	// bind SHMACMemoryInterface
	routerPorts[directionTowardsDest].bind(*dest);
}

void SHMACRouter::connectSHMACMemInterface(shmacsim::direction_enum dir,
		sc_port<SHMACMemoryInterface> & fromRouter,
		sc_port<SHMACMemoryInterface> & toRouter)
{
	toRouter.bind(*this);
	this->routerPorts[dir].bind(fromRouter);
}

void SHMACRouter::connectMaster(SHMACMaster * master)
{
	connectSHMACMemInterface(shmacsim::DIR_INTERN_PROC,
			master->inboundMemoryPort, master->outboundMemoryPort);
}

void SHMACRouter::connectSlave(SHMACSlave* slave)
{
	connectSHMACMemInterface(shmacsim::DIR_INTERN_MEM, slave->inboundMemoryPort,
			slave->outboundMemoryPort);
}
