#include "SHMACRouterArbiter.h"
#include "shmac_typeutils.h"

using namespace shmacsim;

SHMACRouterArbiter::SHMACRouterArbiter(sc_module_name instanceName) :
		sc_module(instanceName)
{
	// TODO reset values - use defines for these instead
	grant_q = emptyDirectionLogicVector;
	pre_req = emptyDirectionLogicVector;
	sel_gnt = emptyDirectionLogicVector;
	isol_lsb = emptyDirectionLogicVector;
	mask_pre = emptyDirectionLogicVector;
	req = emptyDirectionLogicVector;

	SC_METHOD(pickWinners);
	sensitive << grant_q << req << pre_req << mask_pre << isol_lsb << sel_gnt;

	SC_METHOD(convertReqGrantSignals);
	sensitive << grant_q;
	SHMAC_FOREACH_DIR(dir)
		sensitive << req_i[dir];

	SC_CTHREAD(behaviour, clk.pos());
}

void SHMACRouterArbiter::convertReqGrantSignals()
{
	// responsible for two conversion tasks:
	// - copies req bool array -> req logic vector
	// - copies grant logic vector -> grant out dir enum
	// also generates the valid signal, only false when all req & grant = 0

	direction_stdlogic_vector req_v, grant_v;
	grant_v = grant_q.read();
	SHMAC_FOREACH_DIR(dir)
	{
		req_v[dir] = (req_i[dir] ? sc_logic_1 : sc_logic_0);
		if(grant_v[dir] == sc_logic_1)
		{
			//std::cerr << "grant given: " << getDirectionName((shmacsim::direction_enum) dir) << std::endl;
			grant_o = dir;
		}
	}

	req = req_v;

	if((direction_stdlogic_vector)(req_v & grant_v) == emptyDirectionLogicVector)
		valid = false;
	else
		valid = true;
}

void SHMACRouterArbiter::behaviour()
{
	while (1)
	{
		// wait for next positive clock edge
		wait();

		if (enable.read() || grant_q.read().or_reduce() == sc_logic_0)
		{
			pre_req.write(win.read());
			grant_q.write(win.read());
		}
	}
}

void SHMACRouterArbiter::pickWinners()
{
	/*
	 * Translated from the following VHDL snippet:
	 * grant    <= grant_q;
	 mask_pre <=      req and not ( std_logic_vector(unsigned(pre_req) - 1) or pre_req ); -- Mask off previous winners
	 sel_gnt  <= mask_pre and       std_logic_vector(unsigned(not(mask_pre)) + 1);        -- Select new winner
	 isol_lsb <=      req and       std_logic_vector(unsigned(not(req)) + 1);             -- Isolate least significant set bit.
	 win      <= sel_gnt when mask_pre /= (N-1 downto 0 => '0') else isol_lsb;
	 */

	mask_pre = req.read()
			& ~((direction_stdlogic_vector) (pre_req.read().to_uint() - 1)
					| pre_req.read());
	sel_gnt = mask_pre.read()
			& (direction_stdlogic_vector) ((~(mask_pre.read())).to_uint() + 1);
	isol_lsb = req.read()
			& (direction_stdlogic_vector) ((~(req.read())).to_uint() + 1);

	if (mask_pre.read().or_reduce() == sc_logic_0)
		win = isol_lsb.read();
	else
		win = sel_gnt.read();

}
