#include "SHMACRouterCycAcc.h"
#include "shmac_typeutils.h"

SHMACRouterCycAcc::SHMACRouterCycAcc(sc_module_name name, SHMACElement * parent) :
		SHMACRouter(name, parent)
{
	char nameBuffer[128];
	// allocate the router internal structures
	sprintf(nameBuffer, "%s_xb", (const char *) name);
	m_switch = new SHMACRouterCrossbar(nameBuffer);

	// iterate over directions for component allocation
	SHMAC_FOREACH_DIR(dir)
	{
		// allocate the arbiter, input and output port for this direction
		sprintf(nameBuffer, "%s_a(%s)", (const char *) name,
				getDirectionName(dir).c_str());

				m_arbiters[dir] = new SHMACRouterArbiter(nameBuffer);
				sprintf(nameBuffer, "%s_i(%s)", (const char *) name,
						getDirectionName(dir).c_str());

				if (parent)
				m_inputPorts[dir] = new SHMACRouterInputPort(nameBuffer,
						parent->getCoords());
				else
				m_inputPorts[dir] = new SHMACRouterInputPort(nameBuffer,
						{	0, 0});

				sprintf(nameBuffer, "%s_o(%s)", (const char *) name,
						getDirectionName(dir).c_str());
				m_outputPorts[dir] = new SHMACRouterOutputPort(nameBuffer);

				// adapters will be allocated if connectSHMACMemPort* calls are made
				m_adapterIn[dir] = NULL;
				m_adapterOut[dir] = NULL;
	}

	// connect ports with local signals
	SHMAC_FOREACH_DIR(dir)
	{
		// ports to local output signals
		// convenience workaround: the signals bearing outgoing data
		// are allocated from the rx side of the transaction
		// FIXME allocate these on grid-level instead?
		rx_ack[dir].bind(rx_ack_signal[dir]);
		rx_data[dir].bind(rx_data_signal[dir]);
		rx_req[dir].bind(rx_req_signal[dir]);

		// arbiter connections
		m_arbiters[dir]->clk(clk);
		m_arbiters[dir]->enable(m_arbiter_enable[dir]);
		m_arbiters[dir]->valid(m_grant_valid[dir]);
		m_arbiters[dir]->grant_o(m_grant[dir]);

		// output port connections
		m_outputPorts[dir]->clk(clk);
		m_outputPorts[dir]->arbiter_enable(m_arbiter_enable[dir]);
		m_outputPorts[dir]->grant_valid(m_grant_valid[dir]);
		m_outputPorts[dir]->grant(m_grant[dir]);
		m_outputPorts[dir]->router_output_ack(m_router_output_ack[dir]);
		m_outputPorts[dir]->router_output_data(m_router_output_data[dir]);
		m_outputPorts[dir]->tx_ack(tx_ack[dir]);
		m_outputPorts[dir]->tx_data(tx_data[dir]);
		m_outputPorts[dir]->tx_req(tx_req[dir]);

		// input port connections
		m_inputPorts[dir]->clk(clk);
		m_inputPorts[dir]->input_router_ack(m_input_router_ack[dir]);
		m_inputPorts[dir]->input_router_packet(m_input_router_packet[dir]);
		m_inputPorts[dir]->rx_ack(rx_ack[dir]);
		m_inputPorts[dir]->rx_data(rx_data[dir]);
		m_inputPorts[dir]->rx_req(rx_req[dir]);

		// switch connections
		m_switch->rx_ack[dir](m_input_router_ack[dir]);
		m_switch->rx_data[dir](m_input_router_packet[dir]);
		m_switch->tx_ack[dir](m_router_output_ack[dir]);
		m_switch->tx_data[dir](m_router_output_data[dir]);

		// inner iteration over directions to connect remaining signals
		// this is necessary because e.g the requests from a given input port
		// are connected to all possible arbiters in the router
		SHMAC_FOREACH_DIR(dir2)
		{
			// requests
			m_arbiters[dir2]->req_i[dir](m_req[dir][dir2]);
			m_outputPorts[dir2]->router_output_req[dir](m_req[dir][dir2]);
			m_inputPorts[dir]->input_router_req[dir2](m_req[dir][dir2]);
			// grants / "set switch" messages
			m_outputPorts[dir2]->set_switch[dir](m_setsw[dir][dir2]);
			m_switch->arb_grant[dir][dir2](m_setsw[dir][dir2]);
		}

	}

	if (parent)
		setParent(parent);
}

SHMACRouterCycAcc::~SHMACRouterCycAcc()
{
	delete m_switch;
	SHMAC_FOREACH_DIR(dir)
	{
		delete m_arbiters[dir];
		delete m_inputPorts[dir];
		delete m_outputPorts[dir];
		delete m_adapterIn[dir];
		delete m_adapterOut[dir];
	}
}

void SHMACRouterCycAcc::setParent(SHMACElement* parent)
{
	SHMACRouter::setParent(parent);

	SHMAC_FOREACH_DIR(dir)
		m_inputPorts[dir]->setCoords(parent->getCoords());
}

void SHMACRouterCycAcc::connectToRouter(SHMACRouter *dest,
		shmacsim::direction_enum directionTowardsDest)
{
	sc_assert(dest);

	// ensure router models are compatible - cannot e.g connect
	// RTL router model to untimed router model
	if (dest->getRouterModelType() != this->getRouterModelType())
	{
		std::cerr << "Router models are not compatible: " << std::endl
				<< this->getParent()->getCoords() << " has "
				<< this->getRouterModelType() << std::endl
				<< dest->getParent()->getCoords() << " has "
				<< dest->getRouterModelType() << std::endl;
		sc_assert(0);
	}

	// model-specific section: establish connects towards dest
	// connect tx signals of this to rx signals of dest
	shmacsim::direction_enum directionTowardsSrc =
			getDirectionInverse(directionTowardsDest);
	SHMACRouterCycAcc * router2 = (SHMACRouterCycAcc *) dest;

	this->tx_ack[directionTowardsDest].bind(
			router2->rx_ack_signal[directionTowardsSrc]);
	this->tx_data[directionTowardsDest].bind(
			router2->rx_data_signal[directionTowardsSrc]);
	this->tx_req[directionTowardsDest].bind(
			router2->rx_req_signal[directionTowardsSrc]);
}

// workaround for unbound ports
void SHMACRouterCycAcc::plugUnboundTxPorts()
{
	// leaks mem! TODO could be done more cleanly?
	// keep track of alloc'd ptrs & free in destructor
	SHMAC_FOREACH_DIR(dir)
	{
		if (tx_req[dir].bind_count() == 0)
			tx_req[dir].bind(*(new sc_signal<bool>()));

		if (tx_data[dir].bind_count() == 0)
			tx_data[dir].bind(*(new sc_signal<shmacsim::packet>()));

		if (tx_ack[dir].bind_count() == 0)
			tx_ack[dir].bind(*(new sc_signal<bool>()));
	}
}

// SystemC hook - called at the end of the elaboration phase, after the HW is constructed
// we use this opportunity to call plugUnboundTxCalls
void SHMACRouterCycAcc::before_end_of_elaboration()
{
	plugUnboundTxPorts();
}


void SHMACRouterCycAcc::connectSHMACMemInterface(shmacsim::direction_enum dir,
		sc_port<SHMACMemoryInterface> & fromRouter,
		sc_port<SHMACMemoryInterface> & toRouter)
{
	// ensure this dir has not been connected yet
	sc_assert(m_adapterIn[dir] == NULL);
	sc_assert(m_adapterOut[dir] == NULL);

	// allocate adapters
	char nameBuffer[128];
	sprintf(nameBuffer, "%s_%s_i", (const char *) name(), getDirectionName(dir).c_str());
	m_adapterIn[dir] = new SHMACMemToRTLAdapter(nameBuffer);
	sprintf(nameBuffer, "%s_%s_o", (const char *) name(), getDirectionName(dir).c_str());
	m_adapterOut[dir] = new SHMACRTLToMemAdapter(nameBuffer);

	// bind adapter signals to router internal signals
	// adapter -> router
	m_adapterIn[dir]->tx_ack.bind(rx_ack_signal[dir]);
	m_adapterIn[dir]->tx_data.bind(rx_data_signal[dir]);
	m_adapterIn[dir]->tx_req.bind(rx_req_signal[dir]);
	// router -> adapter
	this->tx_ack[dir].bind(m_adapterOut[dir]->rx_ack);
	this->tx_data[dir].bind(m_adapterOut[dir]->rx_data);
	this->tx_req[dir].bind(m_adapterOut[dir]->rx_req);

	// bind SHMACMemoryInterface ports
	// proc -> adapter
	toRouter.bind(*m_adapterIn[dir]);
	// adapter -> proc
	m_adapterOut[dir]->txPort.bind(fromRouter);
}

void sc_trace(sc_trace_file* tf, const SHMACRouterCycAcc& val,
		const std::string& nm)
{
	SHMAC_FOREACH_DIR(dir)
	{
		sc_trace(tf, val.tx_ack[dir], nm + "." + getDirectionName(dir)+ ".tx_ack");

		sc_trace(tf, val.tx_req[dir],
				nm + "." + getDirectionName(dir) + ".tx_req");

		sc_trace(tf, val.tx_data[dir],
				nm + "." + getDirectionName(dir) + ".tx_data");

		sc_trace(tf, val.rx_ack[dir],
				nm + "." + getDirectionName(dir) + ".rx_ack");

		sc_trace(tf, val.rx_req[dir],
				nm + "." + getDirectionName(dir) + ".rx_req");

		sc_trace(tf, val.rx_data[dir],
				nm + "." + getDirectionName(dir) + ".rx_data");
	}
}
