#include "SHMACRouterInputPort.h"
#include "shmac_typeutils.h"
#include "shmac_simutils.h"
#include "SHMACPacketTracker.h"

SHMACRouterInputPort::SHMACRouterInputPort(sc_module_name name,
		shmacsim::tile_coord coord) :
		sc_module(name)
{
	// reset values for internals
	state = next_state = SHMACRouterInputPort::idle;
	input_router_packet_r = dummyPacket;
	m_localCoord = coord;

	SC_METHOD(findRoutingDir);
	sensitive << input_router_packet_i;

	SC_METHOD(fsm);
	sensitive << state << rx_req << rx_data << input_router_ack
			<< input_router_packet_i << input_router_packet_r
			<< routing_direction;

	SC_CTHREAD(sync_proc, clk.pos());
}

void SHMACRouterInputPort::findRoutingDir()
{
	// execute the routing algorithm to find where to send the new packet
	shmacsim::packet message = input_router_packet_i.read();
	shmacsim::direction_enum destDir;

	if (message.req)
		// determine who handles this request next
		destDir = shmacsim::directionOfCoords(m_localCoord,
				getTileCoordFromAddress(message.address));
	else
	{
		// if this is a reply packet for this tile...
		if (message.sender == m_localCoord)
			// special case: override routing destination to INTERN_PROC
			destDir = shmacsim::DIR_INTERN_PROC;
		else
			// replies are sent to the request originator
			destDir = shmacsim::directionOfCoords(m_localCoord, message.sender);
	}

	routing_direction.write(destDir);
}

void SHMACRouterInputPort::sync_proc()
{
	// reset values for externals
	rx_ack = false;
	SHMAC_FOREACH_DIR(dir)
		input_router_req[dir] = false;
	input_router_packet_r = dummyPacket;

	while (1)
	{
		// wait until positive clock edge
		wait();
		// update register values
		state = next_state.read();
		rx_ack = rx_ack_i.read();
		SHMAC_FOREACH_DIR(dir)
			input_router_req[dir] = input_router_req_i[dir].read();
		input_router_packet_r = input_router_packet_i.read();
	}
}

void SHMACRouterInputPort::fsm()
{
	shmacsim::direction_stdlogic_vector oldReqV;
	next_state = state.read();
	rx_ack_i = false;
	SHMAC_FOREACH_DIR(dir)
		input_router_req_i[dir] = false;
	input_router_packet_i = input_router_packet_r.read();
	input_router_packet = input_router_packet_r.read();

	switch (state.read())
	{
	case SHMACRouterInputPort::idle:
		// advance state if outstanding request
		if (rx_req.read())
		{
			next_state = SHMACRouterInputPort::route;
			// annotate timing info if desired
			if (rx_data.read().tag)
				((SHMACPacketTracker *) rx_data.read().tag)->registerLocation(
						"routerIn");
		}
		// receive packet from input
		input_router_packet_i = rx_data.read();
		break;

	case SHMACRouterInputPort::route:
		if (rx_req.read() == false)
			next_state = SHMACRouterInputPort::handshake;
		input_router_req_i[routing_direction.read()] = true;
		// set acknowledge to sender
		rx_ack_i = true;
		break;

	case SHMACRouterInputPort::handshake:
		if (input_router_ack.read())
			next_state = SHMACRouterInputPort::handshake2;
		input_router_req_i[routing_direction.read()] = true;
		// remove ack to sender
		rx_ack_i = false;
		break;

	case SHMACRouterInputPort::handshake2:
		if (input_router_ack.read() == false)
			next_state = SHMACRouterInputPort::idle;
		input_router_req_i[routing_direction.read()] = false;
		break;

	default:
		sc_assert(0);
		// illegal case
		break;
	}
}

