#include "SHMACRouterOutputPort.h"
#include "shmac_typeutils.h"
#include "SHMACPacketTracker.h"

SHMACRouterOutputPort::SHMACRouterOutputPort(sc_module_name name) :
		sc_module(name)
{
	// reset values for internal signals
	router_output_ack_i = false;
	tx_req_i = false;
	arbiter_enable_i = false;
	SHMAC_FOREACH_DIR(dir)
		set_switch_i[dir] = false;

	SC_CTHREAD(sync_proc, clk.pos());

	SC_METHOD(fsm);
	sensitive << state << tx_ack << router_output_data << tx_data_r << grant
			<< grant_valid << grant_r;
	SHMAC_FOREACH_DIR(dir)
		sensitive << router_output_req[dir];
}

void SHMACRouterOutputPort::sync_proc()
{
	// reset values for outputs
	state = SHMACRouterOutputPort::st_idle;
	router_output_ack = false;
	tx_req = false;
	tx_data_r = dummyPacket;
	arbiter_enable = true;
	SHMAC_FOREACH_DIR(dir)
		set_switch[dir] = false;
	grant_r = grant.read();

	while (1)
	{
		wait();
		state = next_state.read();
		router_output_ack = router_output_ack_i.read();
		tx_req = tx_req_i.read();
		tx_data_r = tx_data_i.read();
		arbiter_enable = arbiter_enable_i.read();
		SHMAC_FOREACH_DIR(dir)
			set_switch[dir] = set_switch_i[dir].read();
		grant_r = grant_i.read();
	}
}

void SHMACRouterOutputPort::fsm()
{
	shmacsim::direction_stdlogic_vector switchVector;
	// default values
	next_state = state.read();
	router_output_ack_i = false;
	SHMAC_FOREACH_DIR(dir)
		set_switch_i[dir] = false;
	tx_req_i = false;
	tx_data_i = tx_data_r.read();
	arbiter_enable_i = false;
	grant_i = grant_r.read();

	tx_data = tx_data_r.read();

	switch (state.read())
	{
	case SHMACRouterOutputPort::st_idle:
		if (grant_valid.read())
			next_state = SHMACRouterOutputPort::st_switch;
		arbiter_enable_i = true;
		grant_i = grant.read();
		break;

	case SHMACRouterOutputPort::st_switch:
		next_state = SHMACRouterOutputPort::st_switch_ack;
		// modify desired bit of the switch setting
		set_switch_i[grant_r.read()] = true;
		break;

	case SHMACRouterOutputPort::st_switch_ack:
		next_state = SHMACRouterOutputPort::st_switch_wait;
		// need to read the data at the right moment
		tx_data_i = router_output_data.read();
		// annotate timing if desired
		if (router_output_data.read().tag)
			((SHMACPacketTracker *) router_output_data.read().tag)->registerLocation(
					"routerOut");
		router_output_ack_i = true;
		// modify desired bit of the switch setting
		set_switch_i[grant_r.read()] = true;
		break;

	case SHMACRouterOutputPort::st_switch_wait:
		// not entirely sure this check is necessary, but it exists in output_port.vhd
		if (router_output_req[grant_r.read()].read() == false)
			next_state = SHMACRouterOutputPort::st_send;
		router_output_ack_i = true;
		// modify desired bit of the switch setting
		set_switch_i[grant_r.read()] = true;
		break;

	case SHMACRouterOutputPort::st_send:
		if(tx_ack.read())
			next_state = SHMACRouterOutputPort::st_handshake;
		tx_req_i = true;
		break;

	case SHMACRouterOutputPort::st_handshake:
		if(tx_ack.read() == false)
			next_state = SHMACRouterOutputPort::st_idle;
		tx_data_i = dummyPacket;
		break;

	default:
		sc_assert(0);
		// undefined state
		break;
	}

}
