/*
 * SHMACLockSlave.cpp
 *
 *  Created on: 19 Mar 2013
 *      Author: Yaman Umuroglu
 */

#include "SHMACLockSlave.h"
#include "shmac_simutils.h"

SHMACLockSlave::SHMACLockSlave(sc_module_name name, uint32_t memsize_bytes,
		SHMACElement * parent) :
		SHMACSlave(name, memsize_bytes, parent, false)
{
	// no need to initialize lock table
	// since LL is always called first in a RMW scenario, which initializes

	// initialize lock stats table entries
	for (int i = 0; i < SHMAC_MAX_GRIDDIM * SHMAC_MAX_GRIDDIM; i++)
		m_lockStats[i] =  {	0, 0, 0};
	}

SHMACLockSlave::~SHMACLockSlave()
{
	std::cerr << "Lock tile at " << m_parent->getCoords() << ": statistics"
			<< std::endl;
	std::cerr << "(only non-zero entries shown)" << std::endl;
	std::cerr << "Tile\t#SC-NOK\t\t#SC-OK\t\t#LL"
			<< std::endl;
	for (int i = 0; i < SHMAC_MAX_GRIDDIM * SHMAC_MAX_GRIDDIM; i++)
		if (m_lockStats[i].count_ll)
			std::cerr << shmacsim::coordsFromProcID(i) << "\t"
					<< m_lockStats[i].count_sc_fail << "\t\t"
					<< m_lockStats[i].count_sc_ok << "\t\t"
					<< m_lockStats[i].count_ll << std::endl;
}

// execution context:
// this SHMACLockSlave's own SC_THREAD (performMemoryAccesses)
void SHMACLockSlave::handleMessage(shmacsim::packet message)
{
	// TODO add assertions from SHMACSlave here?
	uint32_t senderID = shmacsim::procIDFromCoords(message.sender);
	shmacsim::packet response;

	/*
	 std::cerr << "sender: " << senderID << " addr " << std::hex
	 << message.address << std::endl;
	 std::cerr << "type "
	 << (message.write_enable == shmacsim::PACKET_WREN_READ4 ?
	 "read" : "write");
	 std::cerr << " entry " << std::hex << m_lockTable[senderID].address
	 << " valid " << m_lockTable[senderID].valid << std::endl
	 << std::endl;
	 */

	// check sync flag to distinguish SC from normal store
	// if sync flag is not set, act like normal RAM
	if (message.flag != shmacsim::PACKET_FLAG_SYNCO)
	{
		SHMACSlave::handleMessage(message);
		return;
	}

	// if we get here, the sync flag is set - LL/SC
	if (message.write_enable == shmacsim::PACKET_WREN_READ4) // load linked (LL)
	{
		shmacsim::waitClockCycles(1);
		// load data normally
		response = SHMACSlave::memoryAccess(message);
		// LL always succeeds, set sync flag on response
		response.flag = shmacsim::PACKET_FLAG_SYNCO;
		// initialize bookkeeping entry for processor
		m_lockTable[senderID].address = message.address;
		m_lockTable[senderID].valid = true;
		// increment LL count for this tile
		m_lockStats[senderID].count_ll++;

		//std::cerr << "LLSC: LL ok, procID " << senderID << " addr " << response.address << " val " << response.data << " at " << sc_time_stamp() << std::endl;

		shmacsim::waitClockCycles(1);
	}
	else  // store conditional (SC)
	{
		shmacsim::waitClockCycles(1);
		if (m_lockTable[senderID].address == message.address
				&& m_lockTable[senderID].valid)
		{
			// SC successful! store the value
			response = SHMACSlave::memoryAccess(message);
			// invalidate all entries for this address
			invalidateLockEntriesForAddress(message.address);
			// set sync flag to indicate success
			response.flag = shmacsim::PACKET_FLAG_SYNCO;
			// increment SC success count for this tile
			m_lockStats[senderID].count_sc_ok++;
			//std::cerr << "LLSC: SC OK, procID " << senderID << " addr " << response.address << " val " << message.data << " at " << sc_time_stamp() << std::endl;
		}
		else
		{
			// SC not successful, don't store anything
			// echo request packet? TODO check if we have any more specific behaviour here
			response = message;
			response.req = false;
			// unset sync flag (if any) to indicate failure
			response.flag = shmacsim::PACKET_FLAG_NONE;
			// increment SC failure count for this tile
			m_lockStats[senderID].count_sc_fail++;
			//std::cerr << "LLSC: SC NOT OK, procID " << senderID  << " addr " << response.address << " val " << message.data << " at " << sc_time_stamp() << std::endl;
		}
		shmacsim::waitClockCycles(1);
	}

	shmacsim::waitClockCycles(1);

	// unlike regular stores, SC needs to send a reply
	// so here we send out the response regardless of load/store flags
	sendMessage(response);
}

void SHMACLockSlave::invalidateLockEntriesForAddress(shmacsim::bus32 addr)
{
	for (uint32_t i = 0; i < SHMAC_MAX_GRIDDIM * SHMAC_MAX_GRIDDIM; i++)
	{
		if (m_lockTable[i].address == addr)
			m_lockTable[i].valid = false;
	}
}

