/*
 * SHMACSlave.cpp
 *
 *  Created on: 26 Feb 2013
 *      Author: Yaman Umuroglu
 */

#include <stdio.h> // for file input-output
#include "SHMACSlave.h"
#include "shmac_typeutils.h"
#include "shmac_simutils.h"
#include "SHMACPacketTracker.h"

SHMACSlave::SHMACSlave(sc_module_name name, uint32_t memsize_bytes,
		SHMACElement * parent, bool readOnly) :
		sc_module(name), m_parent(parent)
{
	m_memory = new uint8_t[memsize_bytes];
	m_readOnly = readOnly;
	m_receivedMessage = dummyPacket;

	inboundMemoryPort.bind(*this);

	SC_THREAD(performMemoryAccesses);

	memset(m_memory, 0, memsize_bytes);

	if (!m_memory)
	{
		std::cerr << "SHMACSlave: could not allocate memory: " << memsize_bytes
				<< " bytes" << std::endl;
		m_allocSize = 0;
	}
	else
		m_allocSize = memsize_bytes;

}

SHMACSlave::~SHMACSlave()
{

	delete[] m_memory;
}

// handle a message sent to this slave in accordance with SHMAC network protocol
void SHMACSlave::receiveMessage(shmacsim::packet message)
{
	m_memLock.lock();
	// 1 cc delay before reading in request
	// memory state machine: idle->ram_wait
	shmacsim::waitClockCycles(1);
	m_receivedMessage = message;
	m_pendingRequest.notify();
	wait(m_requestServed);
	// memory state machine, waiting for clock edges
	shmacsim::waitClockCycles(1);
	m_memLock.unlock();
}

void SHMACSlave::performMemoryAccesses()
{
	while (1)
	{
		// wait until request received
		wait(m_pendingRequest);
		// handle tracker, if exists
		if(m_receivedMessage.tag)
		{
			((SHMACPacketTracker *)m_receivedMessage.tag)->registerLocation("memory");
			// end packet tracking for write pkt
			if(m_receivedMessage.write_enable == shmacsim::PACKET_WREN_WRITE4)
			{
				delete ((SHMACPacketTracker *) m_receivedMessage.tag);
				m_receivedMessage.tag = NULL;
			}
			// other packets tracers will continue on to the reply
		}
		handleMessage(m_receivedMessage);
		m_requestServed.notify();
	}
}

// execution context: this SHMACSlave's own SC_THREAD (performMemoryAccesses)
void SHMACSlave::handleMessage(shmacsim::packet message)
{
	// call the internal handler to carry out the operations
	shmacsim::packet response = memoryAccess(message);

	// send reply only for read operations, no reply needed for stores
	// note that SHMACSlave doesn't care about the sync flag, unlike SHMACLockSlave
	if (message.write_enable == shmacsim::PACKET_WREN_READ4)
		sendMessage(response);
}

shmacsim::packet SHMACSlave::memoryAccess(shmacsim::packet message)
{
	// cannot function without a set parent
	sc_assert(m_parent);
	// a slave can only respond to requests, cannot handle responses
	// TODO fix assertion behaviour: slave response handling
	sc_assert((message.req == true));
	// make sure this message got delivered to the correct location
	tile_coord localCoord = getTileCoordFromAddress(message.address);
	tile_coord myLocalCoord = m_parent->getCoords();
	if(!(localCoord == myLocalCoord))
	{
		std::cerr << "slave " << this->name() << " cannot serve this request:" << message << std::endl;
		sc_assert(0);
	}
	// make sure we can handle the memory range for this packet
	bus32 localAddress = getLocalAddressFromAddress(message.address);
	sc_assert((localAddress < m_allocSize));

	// everything with the packet seems to be in order, handle accordingly
	shmacsim::packet response = message;
	response.req = false;
	response.flag = shmacsim::PACKET_FLAG_NONE;


	if (message.write_enable == shmacsim::PACKET_WREN_READ4)
	{
		// read data from desired location
		response.data = memoryGetWord(localAddress);
	}
	else
	{
		sc_assert(!m_readOnly); // don't write to read only device
		// TODO handle different size writes? only 4 at once for now
		// the ArchC TLM interface only uses 4b writes anyway AFAIK
		memorySetWord(localAddress, message.data);
	}

	return response;
}

// perform word-sized set on slave memory
// internal only, won't respect read-only flag
void SHMACSlave::memorySetWord(SHMAC_ADDR address, SHMAC_WORD data)
{
	//std::cerr << "MEM set \t" << std::hex << address << std::dec << "\t"  << sc_time_stamp() << std::endl;
	// memory state machine: ram_w->write_ack
	shmacsim::waitClockCycles(1);
	*((SHMAC_WORD*) &m_memory[address]) = data;
	// memory state machine: write_ack->idle
	shmacsim::waitClockCycles(1);
}
// perform word-sized get on slave memory
SHMAC_WORD SHMACSlave::memoryGetWord(SHMAC_ADDR address)
{
	//std::cerr << "MEM get \t" << std::hex << address << std::dec << "\t"  << sc_time_stamp() << std::endl;
	// memory state machine: ram_w->read
	shmacsim::waitClockCycles(1);
	// memory state machine: read->read_ack
	shmacsim::waitClockCycles(1);
	// memory state machine: read_ack->idle
	shmacsim::waitClockCycles(1);
	return *((SHMAC_WORD*) &m_memory[address]);
}

// export contents of slave memory to given buffer
void SHMACSlave::exportContents(uint8_t* buffer, uint32_t bufferSize)
{
	memcpy((void *) buffer, (void *) m_memory, bufferSize);
}

// import slave memory content from given buffer
bool SHMACSlave::importContents(uint8_t* buffer, uint32_t bufferSize)
{
	// make sure given buffer fits into slave memory
	if (bufferSize > m_allocSize)
	{
		std::cerr << "SHMACSlave::importContents not enough memory to import"
				<< std::endl;
		return false;
	}

	memcpy((void *) m_memory, (void *) buffer, bufferSize);
	return true;
}

void SHMACSlave::exportContents(const char* fileName)
{
	FILE * exportFile = fopen(fileName, "wb");
	if (!exportFile)
	{
		std::cerr << "SHMACSlave::exportContents could not create file "
				<< fileName << std::endl;
		return;
	}

	fwrite(m_memory, sizeof(uint8_t), m_allocSize, exportFile);
	fclose(exportFile);
}

bool SHMACSlave::importContents(const char* fileName, uint32_t * importedSize)
{
	FILE * importFile = fopen(fileName, "rb");
	if (!importFile)
	{
		std::cerr << "SHMACSlave::importContents could not read from file "
				<< fileName << std::endl;
		return false;
	}

	// get file size
	fseek(importFile, 0, SEEK_END);
	uint32_t fileSize = ftell(importFile);

	if (fileSize > m_allocSize)
	{
		std::cerr
				<< "SHMACSlave::importContents file too big, cannot import. filesize = "
				<< fileSize << " allocsize = " << m_allocSize << std::endl;
		fclose(importFile);
		return false;
	}

	// go back to beginning & read into memory
	fseek(importFile, 0, SEEK_SET);

	if (fread(m_memory, sizeof(uint8_t), fileSize, importFile) != fileSize)
	{
		std::cerr << "SHMACSlave::importContents error reading file "
				<< fileName << std::endl;
		fclose(importFile);
		return false;
	}

	fclose(importFile);

	// pass out imported file size if parameter provided
	if (importedSize)
		*importedSize = fileSize;

	return true;
}

void SHMACSlave::sendMessage(shmacsim::packet message)
{
	//std::cerr << this->name() << ": sending " << message << " at " << sc_time_stamp() << std::endl;
	outboundMemoryPort->receiveMessage(message);
}
