#ifndef UNITTEST_SHMACROUTERCROSSBAR_H_
#define UNITTEST_SHMACROUTERCROSSBAR_H_

#include "router/SHMACRouterCrossbar.h"
#include "shmac_typeutils.h"

class UnitTest_SHMACRouterCrossbar: public sc_module
{
public:
	SC_HAS_PROCESS(UnitTest_SHMACRouterCrossbar);

	sc_signal<bool> arb_grant[SHMAC_DIRECTION_COUNT][SHMAC_DIRECTION_COUNT];

	sc_signal<shmacsim::packet> rx_data[SHMAC_DIRECTION_COUNT];
	sc_signal<bool> rx_ack[SHMAC_DIRECTION_COUNT];

	sc_signal<shmacsim::packet> tx_data[SHMAC_DIRECTION_COUNT];
	sc_signal<bool> tx_ack[SHMAC_DIRECTION_COUNT];

	SHMACRouterCrossbar * uut;
	sc_trace_file* Tf;

	static bool executeUnitTest()
	{
		UnitTest_SHMACRouterCrossbar * unitTest =
				new UnitTest_SHMACRouterCrossbar(
						"UnitTest_SHMACRouterCrossbar");
		// TODO unit test: configure sim duration?
		sc_start(400, SC_NS);

		delete unitTest;
		// TODO unit test: check and report test result
		return true;
	}

	// constructor
	UnitTest_SHMACRouterCrossbar(sc_module_name instanceName) :
			sc_module(instanceName)
	{
		// instantiate the unit to test
		uut = new SHMACRouterCrossbar("uut");

		// set up the trace file
		Tf = sc_create_vcd_trace_file("UnitTest_SHMACRouterCrossbar");
		((vcd_trace_file*) Tf)->set_time_unit(1, SC_NS);

		// add the signals to be traced and bind ports to local signals
		char nmb[15];
		SHMAC_FOREACH_DIR(shmacDir)
		{
			SHMAC_FOREACH_DIR(shmacDir2)
			{
				sprintf(nmb, "arb_grant_%s_%s", getDirectionName(shmacDir).c_str(), getDirectionName(shmacDir2).c_str());
				sc_trace(Tf, arb_grant[shmacDir][shmacDir2], nmb);
				uut->arb_grant[shmacDir][shmacDir2](arb_grant[shmacDir][shmacDir2]);
			}

			sprintf(nmb, "rx_data_%s", getDirectionName(shmacDir).c_str());
			sc_trace(Tf, rx_data[shmacDir], nmb);
			uut->rx_data[shmacDir](rx_data[shmacDir]);

			sprintf(nmb, "rx_ack_%s", getDirectionName(shmacDir).c_str());
			sc_trace(Tf, rx_ack[shmacDir], nmb);
			uut->rx_ack[shmacDir](rx_ack[shmacDir]);

			sprintf(nmb, "tx_data_%s", getDirectionName(shmacDir).c_str());
			sc_trace(Tf, tx_data[shmacDir], nmb);
			uut->tx_data[shmacDir](tx_data[shmacDir]);

			sprintf(nmb, "tx_ack_%s", getDirectionName(shmacDir).c_str());
			sc_trace(Tf, tx_ack[shmacDir], nmb);
			uut->tx_ack[shmacDir](tx_ack[shmacDir]);
		}

		SC_THREAD(run);
	}

	~UnitTest_SHMACRouterCrossbar()
	{
		sc_close_vcd_trace_file(Tf);
		delete uut;
	}

	void clearAllGrantsAndData()
	{
		// clear out all grants, data and acks
		SHMAC_FOREACH_DIR(dir)
		{
			SHMAC_FOREACH_DIR(dir2)
				arb_grant[dir][dir2] = false;
			rx_data[dir] = dummyPacket;
			tx_ack[dir] = false;
		}
	}

	shmacsim::direction_stdlogic_vector onehot_dirv(shmacsim::direction_enum dir)
	{
		shmacsim::direction_stdlogic_vector ret = emptyDirectionLogicVector;
		ret[dir] = sc_logic_1;

		return ret;
	}

	void run()
	{
		shmacsim::packet testPacket;
		testPacket.data = 0xcafebabe;
		testPacket.req = true;
		testPacket.sender =
		{	1,1};
		testPacket.write_enable = shmacsim::PACKET_WREN_WRITE4;
		testPacket.flag = shmacsim::PACKET_FLAG_NONE;
		testPacket.address = 0xdeadbeef;

		// test crossbar action from each direction to each direction
		SHMAC_FOREACH_DIR(dir_src)
		{
			wait(1, SC_NS);

			// for each direction from this direction
			SHMAC_FOREACH_DIR(dir_dest)
			{
				clearAllGrantsAndData();

				// write test packet and grant vector
				rx_data[dir_src].write(testPacket);
				arb_grant[dir_src][dir_dest].write(true);

				wait(1, SC_NS);

				// assert that only desired dest is getting data
				SHMAC_FOREACH_DIR(dir)
					sc_assert(tx_data[dir].read().data == (dir == dir_dest ? 0xcafebabe : 0));

				tx_ack[dir_dest] = true;

				wait(1, SC_NS);

				// assert that only desired src is getting ack
				SHMAC_FOREACH_DIR(dir)
					sc_assert(rx_ack[dir].read() == (dir == dir_src ? true : false));

				std::cout << "test OK: " << getDirectionName(dir_src) << " to " << getDirectionName(dir_dest) << std::endl;

				wait(5, SC_NS);
			}
		}

		// test sending 2 packets simultaneously
		// N to S, E to W
		clearAllGrantsAndData();

		rx_data[shmacsim::DIR_NORTH] = testPacket;
		arb_grant[shmacsim::DIR_NORTH][shmacsim::DIR_SOUTH] = true;

		// alter packet data for E to W
		testPacket.data = 0xbabecafe;
		rx_data[shmacsim::DIR_EAST] = testPacket;

		arb_grant[shmacsim::DIR_EAST][shmacsim::DIR_WEST] = true;

		wait(1, SC_NS);

		sc_assert(tx_data[shmacsim::DIR_SOUTH].read().data == 0xcafebabe);
		sc_assert(tx_data[shmacsim::DIR_WEST].read().data == 0xbabecafe);

		std::cout << "test OK: N to S, E to W" << std::endl;

		std::cout << "Unit test SHMACRouterCrossbar passed!" << std::endl;
	}
};

#endif /* UNITTEST_SHMACROUTERCROSSBAR_H_ */
